
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

/* Thanks to A.Schiffler for his great gfxPrimitives library and to the SGE team
 * many algorithms are based on their work (mainly ellipsis, polygon and beziers)
 */

/*!
 * \file agl_basedraw.c
 * \brief base drawing functions
 * This file implements all the basic drawing functions like lines, rectangles, circles, ellipsis, polygons and beziers. these functions are drawing outlined objects (with and without antialiases).
 * 
 * \author Stephane APIOU
 * \date   mar jui 30 01:33:09 CEST 2002
 */

#include "config.h"
#include "agl_private.h"

#define AGL_DRAWCOLOR		0
#define AGL_DRAWALPHA		1
#define AGL_DRAWAA		2
#define AGL_DRAWPATTERN		3
#define AGL_DRAWFILLEDCOLOR	4
#define AGL_DRAWFILLEDALPHA	5
#define AGL_DRAWFILLEDAA	6
#define AGL_DRAWFILLEDPATTERN	7
#define AGL_DRAWCOUNT		8

#define dopixblend(dst, src) \
	al = ((unsigned char *)(src))[3]; \
	if (al >= 255) \
		*(dst) = *(src); \
	else \
	((unsigned char *)(dst))[0] = (((long)(((unsigned char *)(src))[0] - ((unsigned char *)(dst))[0]) * al) >> 8) + ((unsigned char *)(dst))[0], \
	((unsigned char *)(dst))[1] = (((long)(((unsigned char *)(src))[1] - ((unsigned char *)(dst))[1]) * al) >> 8) + ((unsigned char *)(dst))[1], \
	((unsigned char *)(dst))[2] = (((long)(((unsigned char *)(src))[2] - ((unsigned char *)(dst))[2]) * al) >> 8) + ((unsigned char *)(dst))[2], \
	((unsigned char *)(dst))[3] = (((long)(((unsigned char *)(src))[3] - ((unsigned char *)(dst))[3]) * al) >> 8) + ((unsigned char *)(dst))[3]

#define dopixblendalpha(dst, src, alpha) \
	al = ((unsigned char *)(src))[3] * alpha; \
	if (al >= (255 << 8)) \
		*(dst) = *(src); \
	else \
	((unsigned char *)(dst))[0] = (((long)(((unsigned char *)(src))[0] - ((unsigned char *)(dst))[0]) * al) >> 16) + ((unsigned char *)(dst))[0], \
	((unsigned char *)(dst))[1] = (((long)(((unsigned char *)(src))[1] - ((unsigned char *)(dst))[1]) * al) >> 16) + ((unsigned char *)(dst))[1], \
	((unsigned char *)(dst))[2] = (((long)(((unsigned char *)(src))[2] - ((unsigned char *)(dst))[2]) * al) >> 16) + ((unsigned char *)(dst))[2], \
	((unsigned char *)(dst))[3] = (((long)(((unsigned char *)(src))[3] - ((unsigned char *)(dst))[3]) * al) >> 16) + ((unsigned char *)(dst))[3]


static int
agl_pixelcolor(agl_pix * ptr, agl_pos * pos, unsigned long color)
{
	unsigned long *pixdst;
	short dstrow;

	agl_rvassert(ptr != NULL, -1);

	agl_rvassert(ptr->pixels != NULL, -1);

	agl_rvassert(pos != NULL, -1);

	dstrow = ptr->row_size;

	pixdst = ptr->pixels + pos->y * dstrow + pos->x;
	*pixdst = color;
	return 0;
}

static int
agl_pixelalpha(agl_pix * ptr, agl_pos *pos, unsigned long color)
{
	unsigned long *pixdst;
	short dstrow;
	unsigned long al;

	agl_rvassert(ptr != NULL, -1);

	agl_rvassert(ptr->pixels != NULL, -1);

	agl_rvassert(pos != NULL, -1);

	dstrow = ptr->row_size;

	pixdst = ptr->pixels + pos->y * dstrow + pos->x;
	dopixblend(pixdst, &color);
	return 0;
}

static int
agl_pixelaa(agl_pix * ptr, agl_pos *pos, short bx, short ex, short by, short ey, unsigned long color)
{
	unsigned long *pixdst;
	short dstrow;
	unsigned long al;

	agl_rvassert(ptr != NULL, -1);

	agl_rvassert(ptr->pixels != NULL, -1);

	agl_rvassert(pos != NULL, -1);

	dstrow = ptr->row_size;

	pixdst = ptr->pixels + pos->y * dstrow + pos->x;
	/* do a circle radius 1 around the pixel */
	/* surfaces around the pixel are in ratio of 1 */
	/* .08 | .46 | .08
	 * .46 |  1  | .46
	 * .08 | .46 | .08
	 */
	dopixblend(pixdst, &color);
	if (pos->x > bx) {
		dopixblendalpha(pixdst-1, &color, 116);
		if (pos->y > by)
			dopixblendalpha(pixdst-1-dstrow, &color, 20);
		if ((pos->y+1) < ey)
			dopixblendalpha(pixdst-1+dstrow, &color, 20);
	}
	if ((pos->x+1) < ex) {
		dopixblendalpha(pixdst+1, &color, 116);
		if (pos->y > by)
			dopixblendalpha(pixdst+1-dstrow, &color, 20);
		if ((pos->y+1) < ey)
			dopixblendalpha(pixdst+1+dstrow, &color, 20);
	}
	if (pos->y > by)
		dopixblendalpha(pixdst-dstrow, &color, 116);
	if ((pos->y+1) < ey)
		dopixblendalpha(pixdst+dstrow, &color, 116);
	return 0;
}

int
agl_pixel(agl_pix *ptr, agl_rect *rect, agl_pos *pos, void *type, unsigned long color, unsigned long *ppat)
{
	short nw, nh;
	short dx, dy, ex, ey, dw, dh, bx, by;

	agl_rvassert(ptr != NULL, -1);

	agl_rvassert(ptr->pixels != NULL, -1);

	agl_rvassert(pos != NULL, -1);

	if (rect == NULL)
	{
		dw = ptr->size.width;
		dh = ptr->size.height;
		dx = 0;
		dy = 0;
	}
	else
	{
		dw = rect->width;
		dh = rect->height;
		dx = rect->x;
		dy = rect->y;
	}
	bx = max(0, dx);
	ex = min(ptr->size.width, dx + dw);
	nw = ex - bx;
	by = max(0, dy);
	ey = min(ptr->size.height, dy + dh);
	nh = ey - by;
	if ((nw <= 0) || (nh <= 0))
		return 0;
	if ((pos->x < bx) || (pos->x >= ex) || (pos->y < by) || (pos->y >= ey))
		return 0;
	switch((unsigned long)type) {
		case AGL_DRAWCOLOR:
		case AGL_DRAWFILLEDCOLOR:
			return agl_pixelcolor(ptr, pos, color);
			break;
		case AGL_DRAWALPHA:
		case AGL_DRAWFILLEDALPHA:
			if (AGL_ALPHAVAL(&color) == 255)
				return agl_pixelcolor(ptr, pos, color);
			return agl_pixelalpha(ptr, pos, color);
			break;
		case AGL_DRAWAA:
		case AGL_DRAWFILLEDAA:
			return agl_pixelaa(ptr, pos, bx, ex, by, ey, color);
			break;
		case AGL_DRAWPATTERN:
		case AGL_DRAWFILLEDPATTERN:
			agl_msg("this drawing type is not implemented yet !");
			break;
		case AGL_DRAWCOUNT:
			return 1;
			break;
		default: /* user defined callback line drawing */
			break;
	}
	return -1;
}


//!

/*!
 * \param ptr
 * \param rect
 * \param ipos
 * \param dpos
 * \param  unsignedcolor
 *
 * \return int
*/

static int
agl_linecolor(agl_pix * ptr, agl_pos * ipos, agl_pos * dpos, short by, short ex, short ey, short adx, short ady, unsigned long color)
{
	unsigned long *pixdst;
	short dstrow;
	short cx, cy, ncx, ncy, len, r;
	int i;

	dstrow = ptr->row_size;
	pixdst = ptr->pixels + ipos->y * dstrow + ipos->x;
	if (((ady > 0) && (adx > ady)) || ((ady < 0) && (adx > -ady)))
	{
		if (ipos->y <= dpos->y)
		{
			if (dpos->y < ey)
				ey = dpos->y;
			len = adx / ady;
			r = adx % ady;
			cx = ipos->x;
			ncx = cx + len;
			if (ncx >= ex)
			{
				ncx = ex;
				len = ncx - cx;
			}
			cy = ipos->y;
			while (cy <= ey)
			{
				for (i = 0; i < len; i++)
					*pixdst++ = color;
				cx = ncx;
				if (cx > ex)
					break;
				len = (r + adx) / ady;
				r = (r + adx) % ady;
				ncx = cx + len;
				if (ncx >= ex)
				{
					ncx = ex;
					len = ncx - cx;
				}
				pixdst += dstrow;
				cy++;
			}
		}
		else
		{
			if (dpos->y > by)
				by = dpos->y;
			ady = -ady;
			len = adx / ady;
			r = adx % ady;
			cx = ipos->x;
			ncx = cx + len;
			if (ncx >= ex)
			{
				ncx = ex;
				len = ncx - cx;
			}
			cy = ipos->y;
			while (cy >= by)
			{
				for (i = 0; i < len; i++)
					*pixdst++ = color;
				cx = ncx;
				if (cx > ex)
					break;
				len = (adx + r) / ady;
				r = (adx + r) % ady;
				ncx = cx + len;
				if (ncx >= ex)
				{
					ncx = ex;
					len = ncx - cx;
				}
				cy--;
				pixdst -= dstrow;
			}
		}
	}
	else
	{
		if (ipos->y <= dpos->y)
		{
			if (dpos->x < ex)
				ex = dpos->x;
			len = ady / adx;
			r = ady % adx;
			cy = ipos->y;
			ncy = cy + len;
			if (ncy >= ey)
			{
				ncy = ey;
				len = ncy - cy;
			}
			cx = ipos->x;
			while (cx <= ex)
			{
				for (i = 0; i < len; i++, pixdst += dstrow)
					*pixdst = color;
				cy = ncy;
				if (cy > ey)
					break;
				len = (r + ady) / adx;
				r = (r + ady) % adx;
				ncy = cy + len;
				if (ncy >= ey)
				{
					ncy = ey;
					len = ncy - cy;
				}
				cx++;
				pixdst++;
			}
		}
		else
		{
			if (dpos->x < ex)
				ex = dpos->x;
			ady = -ady;
			len = ady / adx;
			r = ady % adx;
			cy = ipos->y;
			ncy = cy - len;
			if (ncy < by)
			{
				ncy = by;
				len = cy - ncy;
			}
			cx = ipos->x;
			while (cx <= ex)
			{
				for (i = 0; i < len; i++, pixdst -= dstrow)
					*pixdst = color;
				cy = ncy;
				if (cy < by)
					break;
				len = (r + ady) / adx;
				r = (r + ady) % adx;
				ncy = cy - len;
				if (ncy <= by)
				{
					ncy = by;
					len = cy - ncy;
				}
				cx++;
				pixdst++;
			}
		}
	}
	return 1;
}

static int
agl_linealpha(agl_pix * ptr, agl_pos * ipos, agl_pos * dpos, short by, short ex, short ey, short adx, short ady, unsigned long color)
{
	unsigned long *pixdst;
	short dstrow;
	long al;
	short cx, cy, ncx, ncy, len, r;
	int i;

	dstrow = ptr->row_size;
	pixdst = ptr->pixels + ipos->y * dstrow + ipos->x;
	if (((ady > 0) && (adx > ady)) || ((ady < 0) && (adx > -ady)))
	{
		if (ipos->y <= dpos->y)
		{
			if (dpos->y < ey)
				ey = dpos->y;
			len = adx / ady;
			r = adx % ady;
			cx = ipos->x;
			ncx = cx + len;
			if (ncx >= ex)
			{
				ncx = ex;
				len = ncx - cx;
			}
			cy = ipos->y;
			while (cy <= ey)
			{
				for (i = 0; i < len; i++, pixdst++)
					dopixblend(pixdst, &color);
				cx = ncx;
				if (cx > ex)
					break;
				len = (r + adx) / ady;
				r = (r + adx) % ady;
				ncx = cx + len;
				if (ncx >= ex)
				{
					ncx = ex;
					len = ncx - cx;
				}
				pixdst += dstrow;
				cy++;
			}
		}
		else
		{
			if (dpos->y > by)
				by = dpos->y;
			ady = -ady;
			len = adx / ady;
			r = adx % ady;
			cx = ipos->x;
			ncx = cx + len;
			if (ncx >= ex)
			{
				ncx = ex;
				len = ncx - cx;
			}
			cy = ipos->y;
			while (cy >= by)
			{
				for (i = 0; i < len; i++, pixdst++)
					dopixblend(pixdst, &color);
				cx = ncx;
				if (cx > ex)
					break;
				len = (adx + r) / ady;
				r = (adx + r) % ady;
				ncx = cx + len;
				if (ncx >= ex)
				{
					ncx = ex;
					len = ncx - cx;
				}
				cy--;
				pixdst -= dstrow;
			}
		}
	}
	else
	{
		if (ipos->y <= dpos->y)
		{
			if (dpos->x < ex)
				ex = dpos->x;
			len = ady / adx;
			r = ady % adx;
			cy = ipos->y;
			ncy = cy + len;
			if (ncy >= ey)
			{
				ncy = ey;
				len = ncy - cy;
			}
			cx = ipos->x;
			while (cx <= ex)
			{
				for (i = 0; i < len; i++, pixdst += dstrow)
					dopixblend(pixdst, &color);
				cy = ncy;
				if (cy > ey)
					break;
				len = (r + ady) / adx;
				r = (r + ady) % adx;
				ncy = cy + len;
				if (ncy >= ey)
				{
					ncy = ey;
					len = ncy - cy;
				}
				cx++;
				pixdst++;
			}
		}
		else
		{
			if (dpos->x < ex)
				ex = dpos->x;
			ady = -ady;
			len = ady / adx;
			r = ady % adx;
			cy = ipos->y;
			ncy = cy - len;
			if (ncy < by)
			{
				ncy = by;
				len = cy - ncy;
			}
			cx = ipos->x;
			while (cx <= ex)
			{
				for (i = 0; i < len; i++, pixdst -= dstrow)
					dopixblend(pixdst, &color);
				cy = ncy;
				if (cy < by)
					break;
				len = (r + ady) / adx;
				r = (r + ady) % adx;
				ncy = cy - len;
				if (ncy <= by)
				{
					ncy = by;
					len = cy - ncy;
				}
				cx++;
				pixdst++;
			}
		}
	}
	return 1;
}

static int
agl_lineaa(agl_pix * ptr, agl_pos * ipos, agl_pos * dpos, short by, short ex, short ey, short adx, short ady, unsigned long color)
{
	unsigned long *pixdst, *pixnext;
	short dstrow;
	long al;
	short cx, cy, ncx, ncy, len, r;
	int i;
	unsigned char alcur, alnext, aladv;

	dstrow = ptr->row_size;
	pixdst = ptr->pixels + ipos->y * dstrow + ipos->x;
	if ((adx == 0) || (ady == 0))
		return agl_linealpha(ptr, ipos, dpos, by, ex, ey, adx, ady, color);

	if (((ady > 0) && (adx > ady)) || ((ady < 0) && (adx > -ady)))
	{
		if (ipos->y <= dpos->y)
		{
			if (dpos->y < ey)
				ey = dpos->y;
			len = adx / ady;
			r = adx % ady;
			cx = ipos->x;
			ncx = cx + len;
			if (ncx >= ex)
			{
				ncx = ex;
				len = ncx - cx;
			}
			cy = ipos->y;
			alcur = 255;
			alnext = 0;
			aladv = (255 * ady) / adx;
			pixnext = pixdst + dstrow;
			while (cy < ey)
			{
				for (i = 0; i < len; i++, pixdst++, pixnext++) {
					dopixblendalpha(pixdst, &color, alcur);
					dopixblendalpha(pixnext, &color, alnext);
					alcur -= aladv;
					alnext += aladv;
				}
				cx = ncx;
				if (cx > ex)
					break;
				len = (r + adx) / ady;
				r = (r + adx) % ady;
				ncx = cx + len;
				if (ncx >= ex)
				{
					ncx = ex;
					len = ncx - cx;
				}
				alcur = 255;
				alnext = 0;
				aladv = (255 * ady) / (adx + r);
				pixdst = pixnext;
				pixnext += dstrow;
				cy++;
			}
		}
		else
		{
			if (dpos->y > by)
				by = dpos->y;
			ady = -ady;
			len = adx / ady;
			r = adx % ady;
			cx = ipos->x;
			ncx = cx + len;
			if (ncx >= ex)
			{
				ncx = ex;
				len = ncx - cx;
			}
			cy = ipos->y;
			alcur = 255;
			alnext = 0;
			aladv = (255 * ady) / adx;
			pixnext = pixdst - dstrow;
			while (cy > by)
			{
				for (i = 0; i < len; i++, pixdst++, pixnext++) {
					dopixblendalpha(pixdst, &color, alcur);
					dopixblendalpha(pixnext, &color, alnext);
					alcur -= aladv;
					alnext += aladv;
				}
				cx = ncx;
				if (cx > ex)
					break;
				len = (adx + r) / ady;
				r = (adx + r) % ady;
				ncx = cx + len;
				if (ncx >= ex)
				{
					ncx = ex;
					len = ncx - cx;
				}
				alcur = 255;
				alnext = 0;
				aladv = (255 * ady) / (adx + r);
				pixdst = pixnext;
				pixnext -= dstrow;
				cy--;
			}
		}
	}
	else
	{
		if (ipos->y <= dpos->y)
		{
			if (dpos->x < ex)
				ex = dpos->x;
			len = ady / adx;
			r = ady % adx;
			cy = ipos->y;
			ncy = cy + len;
			if (ncy >= ey)
			{
				ncy = ey;
				len = ncy - cy;
			}
			cx = ipos->x;
			alcur = 255;
			alnext = 0;
			aladv = (255 * adx) / ady;
			while (cx < ex)
			{
				for (i = 0; i < len; i++, pixdst += dstrow) {
					dopixblendalpha(pixdst, &color, alcur);
					dopixblendalpha(pixdst + 1, &color, alnext);
					alcur -= aladv;
					alnext += aladv;
				}
				cy = ncy;
				if (cy > ey)
					break;
				len = (r + ady) / adx;
				r = (r + ady) % adx;
				ncy = cy + len;
				if (ncy >= ey)
				{
					ncy = ey;
					len = ncy - cy;
				}
				cx++;
				pixdst++;
				alcur = 255;
				alnext = 0;
				aladv = (255 * ady) / (adx + r);
			}
		}
		else
		{
			if (dpos->x < ex)
				ex = dpos->x;
			ady = -ady;
			len = ady / adx;
			r = ady % adx;
			cy = ipos->y;
			ncy = cy - len;
			if (ncy < by)
			{
				ncy = by;
				len = cy - ncy;
			}
			cx = ipos->x;
			alcur = 255;
			alnext = 0;
			aladv = (255 * adx) / ady;
			while (cx < ex)
			{
				for (i = 0; i < len; i++, pixdst -= dstrow) {
					dopixblendalpha(pixdst, &color, alcur);
					dopixblendalpha(pixdst + 1, &color, alnext);
					alcur -= aladv;
					alnext += aladv;
				}
				cy = ncy;
				if (cy < by)
					break;
				len = (r + ady) / adx;
				r = (r + ady) % adx;
				ncy = cy - len;
				if (ncy <= by)
				{
					ncy = by;
					len = cy - ncy;
				}
				cx++;
				pixdst++;
				alcur = 255;
				alnext = 0;
				aladv = (255 * ady) / (adx + r);
			}
		}
	}
	return 1;
}

static int
agl_linecount(agl_pos * ipos, agl_pos * dpos, short by, short ex, short ey, short adx, short ady)
{
	short cx, cy, ncx, ncy, len, r;
	int count;

	count =0;
	if (((ady > 0) && (adx > ady)) || ((ady < 0) && (adx > -ady)))
	{
		if (ipos->y <= dpos->y)
		{
			if (dpos->y < ey)
				ey = dpos->y;
			len = adx / ady;
			r = adx % ady;
			cx = ipos->x;
			ncx = cx + len;
			if (ncx >= ex)
			{
				ncx = ex;
				len = ncx - cx;
			}
			cy = ipos->y;
			while (cy <= ey)
			{
				count += len;
				cx = ncx;
				if (cx > ex)
					break;
				len = (r + adx) / ady;
				r = (r + adx) % ady;
				ncx = cx + len;
				if (ncx >= ex)
				{
					ncx = ex;
					len = ncx - cx;
				}
				cy++;
			}
		}
		else
		{
			if (dpos->y > by)
				by = dpos->y;
			ady = -ady;
			len = adx / ady;
			r = adx % ady;
			cx = ipos->x;
			ncx = cx + len;
			if (ncx >= ex)
			{
				ncx = ex;
				len = ncx - cx;
			}
			cy = ipos->y;
			while (cy >= by)
			{
				count += len;
				cx = ncx;
				if (cx > ex)
					break;
				len = (adx + r) / ady;
				r = (adx + r) % ady;
				ncx = cx + len;
				if (ncx >= ex)
				{
					ncx = ex;
					len = ncx - cx;
				}
				cy--;
			}
		}
	}
	else
	{
		if (ipos->y <= dpos->y)
		{
			if (dpos->x < ex)
				ex = dpos->x;
			len = ady / adx;
			r = ady % adx;
			cy = ipos->y;
			ncy = cy + len;
			if (ncy >= ey)
			{
				ncy = ey;
				len = ncy - cy;
			}
			cx = ipos->x;
			while (cx <= ex)
			{
				count += len;
				cy = ncy;
				if (cy > ey)
					break;
				len = (r + ady) / adx;
				r = (r + ady) % adx;
				ncy = cy + len;
				if (ncy >= ey)
				{
					ncy = ey;
					len = ncy - cy;
				}
				cx++;
			}
		}
		else
		{
			if (dpos->x < ex)
				ex = dpos->x;
			ady = -ady;
			len = ady / adx;
			r = ady % adx;
			cy = ipos->y;
			ncy = cy - len;
			if (ncy < by)
			{
				ncy = by;
				len = cy - ncy;
			}
			cx = ipos->x;
			while (cx <= ex)
			{
				count += len;
				cy = ncy;
				if (cy < by)
					break;
				len = (r + ady) / adx;
				r = (r + ady) % adx;
				ncy = cy - len;
				if (ncy <= by)
				{
					ncy = by;
					len = cy - ncy;
				}
				cx++;
			}
		}
	}
	return count;
}

int
agl_line(agl_pix *ptr, agl_rect *rect, agl_pos *ipos, agl_pos *dpos, void *type, unsigned long color, unsigned long *ppat, unsigned short width)
{
	short nw, nh, tmp;
	short dx, dy, ex, ey, dw, dh, bx, by;
	short adx, ady;
	long trans;
	int k;
	agl_pos ilpos, dlpos;

	agl_rvassert(ptr != NULL, -1);

	agl_rvassert(ptr->pixels != NULL, -1);

	agl_rvassert(ipos != NULL, -1);

	agl_rvassert(dpos != NULL, -1);

	if (width <= 1) {
		if (rect == NULL)
		{
			dw = ptr->size.width;
			dh = ptr->size.height;
			dx = 0;
			dy = 0;
		}
		else
		{
			dw = rect->width;
			dh = rect->height;
			dx = rect->x;
			dy = rect->y;
		}
		bx = max(0, dx);
		ex = min(ptr->size.width, dx + dw);
		nw = ex - bx;
		by = max(0, dy);
		ey = min(ptr->size.height, dy + dh);
		nh = ey - by;
		if ((nw <= 0) || (nh <= 0))
			return 0;
		if (ipos->x > dpos->x)
		{
			tmp = ipos->x;
			ipos->x = dpos->x;
			dpos->x = tmp;
			tmp = ipos->y;
			ipos->y = dpos->y;
			dpos->y = tmp;
		}
		if (dpos->x <= bx)
			return 0;
		if (ipos->x >= ex)
			return 0;
		if (ipos->y < dpos->y)
		{
			if ((ipos->y >= ey) || (dpos->y <= by))
				return 0;
		}
		else
		{
			if ((dpos->y >= ey) || (ipos->y <= by))
				return 0;
		}
		adx = dpos->x - ipos->x + 1;
		ady = dpos->y - ipos->y;
		ady += (ady < 0)? -1:1;
		trans = ady * (bx - ipos->x) - adx * (by - ipos->y);
		if (ipos->x > bx)
		{
			if (ipos->y <= by)
			{
				ipos->x += ((by - ipos->y) * adx) / ady;
				if (ipos->x >= ex)
					return 0;
				ipos->y = by;
			}
			else if (ipos->y >= ey)
			{
				ipos->x += ((ey - ipos->y) * adx) / ady;
				if (ipos->x >= ex)
					return 0;
				ipos->y = ey;
			}

		}
		else
		{
			if (ipos->y <= by)
			{
				if (trans > 0)
				{
					ipos->x += ((by - ipos->y) * adx) / ady;
					if (ipos->x >= ex)
						return 0;
					ipos->y = by;
				}
				else
				{
					ipos->y = (ady * (bx - ipos->x)) / adx;
					if ((ipos->y <= by) || (ipos->y >= ey))
						return 0;
					ipos->x = bx;
				}
			}
			else if (ipos->y >= ey)
			{
				if (trans > (nh * adx))
				{
					ipos->y = (ady * (bx - ipos->x)) / adx;
					if ((ipos->y <= by) || (ipos->y >= ey))
						return 0;
					ipos->x = bx;
				}
				else
				{
					ipos->x += ((ey - ipos->y) * adx) / ady;
					if (ipos->x >= ex)
						return 0;
					ipos->y = ey;
				}
			}
			else
			{
				ipos->y = (ady * (bx - ipos->x)) / adx;
				if ((ipos->y <= by) || (ipos->y >= ey))
					return 0;
				ipos->x = bx;
			}
		}
		switch((unsigned long)type) {
		case AGL_DRAWCOLOR:
			return agl_linecolor(ptr, ipos, dpos, by, ex, ey, adx, ady, color);
			break;
		case AGL_DRAWALPHA:
			if (AGL_ALPHAVAL(&color) == 255)
				return agl_linecolor(ptr, ipos, dpos, by, ex, ey, adx, ady, color);
			return agl_linealpha(ptr, ipos, dpos, by, ex, ey, adx, ady, color);
			break;
		case AGL_DRAWAA:
			return agl_lineaa(ptr, ipos, dpos, by, ex, ey, adx, ady, color);
			break;
		case AGL_DRAWPATTERN:
			break;
		case AGL_DRAWCOUNT:
			return agl_linecount(ipos, dpos, by, ex, ey, adx, ady);
			break;
		default: /* user defined callback line drawing */
			break;
		}
	} else {
		if (rect == NULL)
		{
			dw = ptr->size.width;
			dh = ptr->size.height;
			dx = 0;
			dy = 0;
		}
		else
		{
			dw = rect->width;
			dh = rect->height;
			dx = rect->x;
			dy = rect->y;
		}
		bx = max(0, dx);
		ex = min(ptr->size.width, dx + dw);
		nw = ex - bx;
		by = max(0, dy);
		ey = min(ptr->size.height, dy + dh);
		nh = ey - by;
		if ((nw <= 0) || (nh <= 0))
			return 0;
		if (ipos->x > dpos->x)
		{
			tmp = ipos->x;
			ipos->x = dpos->x;
			dpos->x = tmp;
			tmp = ipos->y;
			ipos->y = dpos->y;
			dpos->y = tmp;
		}
		for (k=0; k < width; k++) {
			ilpos.x= ipos->x;
			ilpos.y= ipos->y+k;
			dlpos.x= dpos->x;
			dlpos.y= dpos->y+k;
			if (dlpos.x <= bx)
				continue;
			if (ilpos.x >= ex)
				continue;
			if (ilpos.y < dlpos.y)
			{
				if ((ilpos.y >= ey) || (dlpos.y <= by))
					continue;
			}
			else
			{
				if ((dlpos.y >= ey) || (ilpos.y <= by))
					continue;
			}
			adx = dlpos.x - ilpos.x + 1;
			ady = dlpos.y - ilpos.y;
			ady += (ady < 0)? -1:1;
			trans = ady * (bx - ilpos.x) - adx * (by - ilpos.y);
			if (ilpos.x > bx)
			{
				if (ilpos.y <= by)
				{
					ilpos.x += ((by - ilpos.y) * adx) / ady;
					if (ilpos.x >= ex)
						continue;
					ilpos.y = by;
				}
				else if (ilpos.y >= ey)
				{
					ilpos.x += ((ey - ilpos.y) * adx) / ady;
					if (ilpos.x >= ex)
						continue;
					ilpos.y = ey;
				}

			}
			else
			{
				if (ilpos.y <= by)
				{
					if (trans > 0)
					{
						ilpos.x += ((by - ilpos.y) * adx) / ady;
						if (ilpos.x >= ex)
							continue;
						ilpos.y = by;
					}
					else
					{
						ilpos.y = (ady * (bx - ilpos.x)) / adx;
						if ((ilpos.y <= by) || (ilpos.y >= ey))
							continue;
						ilpos.x = bx;
					}
				}
				else if (ilpos.y >= ey)
				{
					if (trans > (nh * adx))
					{
						ilpos.y = (ady * (bx - ilpos.x)) / adx;
						if ((ilpos.y <= by) || (ilpos.y >= ey))
							continue;
						ilpos.x = bx;
					}
					else
					{
						ilpos.x += ((ey - ilpos.y) * adx) / ady;
						if (ilpos.x >= ex)
							continue;
						ilpos.y = ey;
					}
				}
				else
				{
					ilpos.y = (ady * (bx - ilpos.x)) / adx;
					if ((ilpos.y <= by) || (ilpos.y >= ey))
						continue;
					ilpos.x = bx;
				}
			}
			switch((unsigned long)type) {
			case AGL_DRAWCOLOR:
			case AGL_DRAWFILLEDCOLOR:
				if (agl_linecolor(ptr, &ilpos, &dlpos, by, ex, ey, adx, ady, color) < 0)
					return -1;
				break;
			case AGL_DRAWALPHA:
			case AGL_DRAWFILLEDALPHA:
				if (AGL_ALPHAVAL(&color) == 255) {
					if (agl_linecolor(ptr, &ilpos, &dlpos, by, ex, ey, adx, ady, color) < 0)
						return -1;
				} else
					if (agl_linealpha(ptr, &ilpos, &dlpos, by, ex, ey, adx, ady, color) < 0)
						return -1;
				break;
			case AGL_DRAWAA:
			case AGL_DRAWFILLEDAA:
				if (agl_lineaa(ptr, &ilpos, &dlpos, by, ex, ey, adx, ady, color) < 0)
					return -1;
				break;
			case AGL_DRAWPATTERN:
			case AGL_DRAWFILLEDPATTERN:
				break;
			case AGL_DRAWCOUNT:
				return agl_linecount(&ilpos, &dlpos, by, ex, ey, adx, ady);
				break;
				default: /* user defined callback line drawing */
					break;
			}
		}
	}
	return -1;
}

int
agl_rectangle(agl_pix *ptr, agl_rect *rect, agl_pos *ipos, agl_pos *dpos, void *type, unsigned long color, unsigned long *ppat, unsigned short width)
{
	short nw, nh, tmp;
	short dx, dy, ex, ey, dw, dh, bx, by;
	short adx, ady;
	long trans;
	int k;

	agl_rvassert(ptr != NULL, -1);

	agl_rvassert(ptr->pixels != NULL, -1);

	agl_rvassert(ipos != NULL, -1);

	agl_rvassert(dpos != NULL, -1);

	if (rect == NULL)
	{
		dw = ptr->size.width;
		dh = ptr->size.height;
		dx = 0;
		dy = 0;
	}
	else
	{
		dw = rect->width;
		dh = rect->height;
		dx = rect->x;
		dy = rect->y;
	}
	bx = max(0, dx);
	ex = min(ptr->size.width, dx + dw);
	nw = ex - bx;
	by = max(0, dy);
	ey = min(ptr->size.height, dy + dh);
	nh = ey - by;
	if ((nw <= 0) || (nh <= 0))
		return 0;
	if (ipos->x > dpos->x)
	{
		tmp = ipos->x;
		ipos->x = dpos->x;
		dpos->x = tmp;
		tmp = ipos->y;
		ipos->y = dpos->y;
		dpos->y = tmp;
	}
	if (dpos->x <= bx)
		return 0;
	if (ipos->x >= ex)
		return 0;
	if (ipos->y < dpos->y)
	{
		if ((ipos->y >= ey) || (dpos->y <= by))
			return 0;
	}
	else
	{
		if ((dpos->y >= ey) || (ipos->y <= by))
			return 0;
	}
	adx = dpos->x - ipos->x + 1;
	ady = dpos->y - ipos->y;
	ady += (ady < 0)? -1:1;
	trans = ady * (bx - ipos->x) - adx * (by - ipos->y);
	if (ipos->x > bx)
	{
		if (ipos->y <= by)
		{
			ipos->x += ((by - ipos->y) * adx) / ady;
			if (ipos->x >= ex)
				return 0;
			ipos->y = by;
		}
		else if (ipos->y >= ey)
		{
			ipos->x += ((ey - ipos->y) * adx) / ady;
			if (ipos->x >= ex)
				return 0;
			ipos->y = ey;
		}

	}
	else
	{
		if (ipos->y <= by)
		{
			if (trans > 0)
			{
				ipos->x += ((by - ipos->y) * adx) / ady;
				if (ipos->x >= ex)
					return 0;
				ipos->y = by;
			}
			else
			{
				ipos->y = (ady * (bx - ipos->x)) / adx;
				if ((ipos->y <= by) || (ipos->y >= ey))
					return 0;
				ipos->x = bx;
			}
		}
		else if (ipos->y >= ey)
		{
			if (trans > (nh * adx))
			{
				ipos->y = (ady * (bx - ipos->x)) / adx;
				if ((ipos->y <= by) || (ipos->y >= ey))
					return 0;
				ipos->x = bx;
			}
			else
			{
				ipos->x += ((ey - ipos->y) * adx) / ady;
				if (ipos->x >= ex)
					return 0;
				ipos->y = ey;
			}
		}
		else
		{
			ipos->y = (ady * (bx - ipos->x)) / adx;
			if ((ipos->y <= by) || (ipos->y >= ey))
				return 0;
			ipos->x = bx;
		}
	}
	switch((unsigned long)type) {
		case AGL_DRAWCOLOR:
			return agl_linecolor(ptr, ipos, dpos, by, ex, ey, adx, ady, color);
			break;
		case AGL_DRAWALPHA:
			if (AGL_ALPHAVAL(&color) == 255)
				return agl_linecolor(ptr, ipos, dpos, by, ex, ey, adx, ady, color);
			return agl_linealpha(ptr, ipos, dpos, by, ex, ey, adx, ady, color);
			break;
		case AGL_DRAWAA:
			return agl_lineaa(ptr, ipos, dpos, by, ex, ey, adx, ady, color);
			break;
		case AGL_DRAWPATTERN:
			break;
		case AGL_DRAWCOUNT:
			return agl_linecount(ipos, dpos, by, ex, ey, adx, ady);
			break;
		default: /* user defined callback line drawing */
			break;
	}
	return -1;
}
